#include <nemo/str.h>

#include "mess822.h"

static unsigned int state; /* 0: no address; 1: address without @; 2: address with @ */
static sa_vector addr = SA_VECTOR;
static sa_vector comment = SA_VECTOR;

static stralloc tmp = STRALLOC;

static unsigned int do_comment(mess822_addrlist *out)
{
  unsigned int i;
  unsigned int start;
  unsigned int end;
  char prev_type;
  char type;

  if (!(end = comment.len)) return 1;  /* empty list */
/*
  find last non-blank in list
*/
  while (end) {
    if (comment.va[end - 1].s[0] != ' ') break;
    end--;
  }
/*
  find first non-blank in list
*/
  start = 0;
  while (start < end) {
    if (comment.va[start].s[0] != ' ') break;
    start++;
  }
  comment.len = 0;

  if (start == end) return 1;  /* list only contains white space */

  if (!stralloc_erase(&tmp)) return 0;
  i = end;
  type = ' ';
  do {
    i--;
    prev_type = type;
    type = comment.va[i].s[0];
    if (type == ' ' && prev_type == ' ') continue;
    if (!stralloc_cat(&tmp, &comment.va[i])) return 0;
  } while (i > start);

  return mess822_addrlist_append(out, &tmp, '(');
}

static int do_it(mess822_addrlist *out)
{
  unsigned int i;

  if (!state) return 1;  /* not address */

  if (!stralloc_erase(&tmp)) return 0;

  i = addr.len;
  while (i) {  /* reverse order */
    if (!stralloc_cat(&tmp, &addr.va[--i])) return 0;
  }
  if (state == 1) {  /* missing '@', so add one */
    if (!stralloc_cats(&tmp, "@")) return 0;
  }
  if (!mess822_addrlist_append(out, &tmp, '@')) return 0;

  if (!do_comment(out)) return 0;

  state = 0;
  addr.len = 0;
  return 1;
}

static unsigned int add_comment_blank(void)
{
  if (!stralloc_copys(&tmp, " ")) return 0;
  return sa_vector_append(&comment, &tmp);
}

static unsigned int add_comment(const stralloc *token)
{
  unsigned int start;
  char token_type;

  token_type = token->s[0];

  if (token_type == ',') return 1;
  if (token_type == ':') return 1;
  if (token_type == ';') return 1;
  if (token_type == '<') return 1;
  if (token_type == '>') return 1;
  if (token_type == '=') return 1;

  start = 0;
  if ((token_type == '%') || (token_type == '(')) {
    ++start;
  }

  if (!stralloc_copyb(&tmp, token->s + start, token->len - start)) return 0;
  return sa_vector_append(&comment, &tmp);
}

static unsigned int add_address(const stralloc *token)
{
  unsigned int start;
  char token_type;

  token_type = token->s[0];
  if ((token_type != '%') && (token_type != '.') && (token_type != '@')) {
    return add_comment(token);
  }

  if (!state) {
    state = 1;
  }
  if (token_type == '@') {
    state = 2;
  }

  start = 0;
  if (token_type == '%') {
    ++start;
  }

  if (!stralloc_copyb(&tmp, token->s + start, token->len - start)) return 0;

  return sa_vector_append(&addr, &tmp);
}

unsigned int mess822_addrlist_scan(mess822_addrlist *out, const stralloc *in)
{
  static sa_vector tokens = SA_VECTOR;
  static mess822_addrlist list = MESS822_ADDRLIST;

  stralloc *token;
  mess822_addrdata *addrdata;
  unsigned int flag_word_ok;
  unsigned int flag_phrase;
  unsigned int j;
  char token_type;

  if (!mess822_token(&tokens, in)) return 0;
  if (!mess822_addrlist_erase(out)) return 0;
  if (!mess822_addrlist_erase(&list)) return 0;
  if (!sa_vector_erase(&comment)) return 0;
  if (!sa_vector_erase(&addr)) return 0;

  flag_word_ok = 1;
  flag_phrase = 0;
  state = 0;

  j = tokens.len;

  while (j) {
    j--;
    token = &tokens.va[j];
    token_type = token->s[0];

    if (flag_phrase) {
      if ((token_type != ',') && (token_type != ';') && (token_type != ':') && (token_type != '>')) {
        if (!add_comment(token)) return 0;
        continue;
      }
      if (!do_it(&list)) return 0;
      flag_phrase = 0;
      flag_word_ok = 1;
    }

    switch (token_type) {
      case ' ':
      case '\t':
        if (!add_comment_blank()) return 0;
        break;

      case ';':
        if (!do_it(&list)) return 0;
        if (!do_comment(&list)) return 0;
        if (!stralloc_copys(&tmp, ";")) return 0;
        if (!mess822_addrlist_append(&list, &tmp, ';')) return 0;
        flag_word_ok = 1;
        break;

      case ',':
        if (!do_it(&list)) return 0;
        if (!do_comment(&list)) return 0;
        flag_word_ok = 1;
        break;

      case '(':
        if (!add_comment_blank()) return 0;
        if (!add_comment(token)) return 0;
        if (!add_comment_blank()) return 0;
        break;

      case '%':
        if (!flag_word_ok) {
          if (!do_it(&list)) return 0;
        }
        if (!add_address(token)) return 0;
        flag_word_ok = 0;
        break;

      case '>':
        if (!do_it(&list)) return 0;

        if (!state) state = 1;  /* <> is an address */

        for (;;) {
          if (!add_address(token)) return 0;
          if (!j) break;
          j--;
	  token = &tokens.va[j];
	  token_type = token->s[0];
          if (token_type == ':') break;
          if (token_type == '<') break;
        }

        if (token_type == ':') {
          for (;;) {
            if (!add_comment(token)) return 0;
            if (!j) break;
            j--;
            token = &tokens.va[j];
	    token_type = token->s[0];
            if (token_type == '<') break;
          }
        }
        flag_phrase = 1;
        break;

      case ':':
        if (!do_it(&list)) return 0;
        if (!do_comment(&list)) return 0;
        if (!stralloc_copys(&tmp, ":")) return 0;
        if (!mess822_addrlist_append(&list, &tmp, ':')) return 0;
        flag_phrase = 1;
        break;

      default:
        if (!add_address(token)) return 0;
        flag_word_ok = 1;
        break;
    }
  }

  if (!do_it(&list)) return 0;
  if (!do_comment(&list)) return 0;
/*
  reverse order of items
*/
  j = list.len;
  while (j) {
    addrdata = &list.ad[--j];
    if (!mess822_addrlist_append(out, &addrdata->data, addrdata->type)) return 0;
  }

  return 1;
}
